// ----------------------------------------------------------------------
// StartUp! - About window
//
// Copyright © 2000 Arcane Dragon Software (arcanedragon@pobox.com)
//
// Display and manage the About window.
// ----------------------------------------------------------------------

#include "AboutWindow.h"
#include "App.h"
#include "AppIconView.h"
#include "StartupSettings.h"
#include "URLStringView.h"

#include <game/FileGameSound.h>
#include <interface/Font.h>
#include <storage/AppFileInfo.h>
#include <storage/Resources.h>
#include <stdio.h>
#include <string.h>

#include "HGroup.h"
#include "MBViewWrapper.h"
#include "MBorder.h"
#include "MStringView.h"
#include "Space.h"
#include "VGroup.h"

// +----------------------+
// | [ ]  About StartUp!… |
// +----------------------+-----------------------------+
// |                                                    |
// |  +--+                                              |
// |  |1 |  StartUp! x.y.z (type/build)                 |
// |  +--+                                              |
// |        Copyright © 2000 Arcane Dragon Softfware    |
// |                        (arcanedragon@pobox.com)    |
// |                                                    |
// |         Design and development - Chris Herborth    |
// |                                                    |
// |           liblayout UI library - Marco Nelisson    |
// |       Santa's Gift Bag library - Brian Tietz       |
// |                  Sound effects - Partners in Rhyme |
// |                                                    |
// |                        Support - NOT REGISTERED    |
// |                                  NOT REGISTERED    |
// |                                                    |
// +----------------------------------------------------+

// ----------------------------------------------------------------------
// Local constants

static const char *SETTING_POSITION = "about:position";

// ----------------------------------------------------------------------
// Constructor/destructor

AboutWindow::AboutWindow( void )
	: MWindow( BRect( 100, 100, 101, 101 ), 
			   static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "about title", &_size ) ),
			   B_TITLED_WINDOW_LOOK, B_NORMAL_WINDOW_FEEL,
			   B_NOT_RESIZABLE | B_NOT_ZOOMABLE |
//			   B_WILL_ACCEPT_FIRST_CLICK | B_ASYNCHRONOUS_CONTROLS ),
			   B_ASYNCHRONOUS_CONTROLS ),
	  _sfx( NULL )
{
	App *app = static_cast<App *>( be_app );
	app->RegisterWindow( this );

	// Load strings from resources...
	BResources *rsrc = be_app->AppResources();

	const char *about_title_format = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about title format", &_size ) );
	const char *about_copyright = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about copyright", &_size ) );
	const char *about_email = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "ADS email", &_size ) );
	const char *about_design = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about design", &_size ) );
	const char *about_liblayout = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about liblayout", &_size ) );
	const char *about_sgb = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about SGB", &_size ) );
	const char *about_sfx = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about sfx", &_size ) );
#if defined(SHAREWARE_NAG)
	const char *about_user = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about user", &_size ) );
#endif
	const char *about_emdash = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "em-dash", &_size ) );
	const char *about_designer = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about designer", &_size ) );
	const char *about_liblayout_author = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about liblayout author", &_size ) );
	const char *about_sgb_author = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about SGB author", &_size ) );
	const char *about_sfx_author = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about sfx author", &_size ) );
	const char *about_liblayout_url = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "liblayout URL", &_size ) );
	const char *about_sgb_url = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "SGB URL", &_size ) );
	const char *about_sfx_url = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "Partners In Rhyme URL", &_size ) );

	StartupSettings *settings = app->GetSettings();
#if defined(SHAREWARE_NAG)
	const char *about_reg_name = NULL;
	const char *about_reg_company = NULL;
	if( settings->FindString( "reg:name", &about_reg_name ) != B_OK ) {
		about_reg_name = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "not registered", &_size ) );
	}
	if( settings->FindString( "reg:company", &about_reg_company ) != B_OK ) {
		about_reg_company = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "not registered", &_size ) );
	}
#endif

	// Fill in the application title info with some version information.
	char app_version_text[256];
	version_info vi;
	BAppFileInfo *appi = app->GetInfo();
	status_t retval = appi->GetVersionInfo( &vi, B_APP_VERSION_KIND );
	if( retval != B_OK ) memset( &vi, 0, sizeof( version_info ) );

	snprintf( app_version_text, 256, about_title_format,
			  vi.major, vi.middle, vi.minor,
			  VersionVariety( vi.variety ), vi.internal );

	// Views we'll want to mess with later...
	MStringView *title = NULL;
	URLStringView *email = NULL;
	MStringView *chrish = NULL;
	MStringView *marcone = NULL;
	MStringView *bteitz = NULL;
	MStringView *pir = NULL;
#if defined(SHAREWARE_NAG)
	MStringView *reg_name = NULL;
#endif

	// Build the window; this unholy mess doesn't quite give me the results
	// I want.  It's difficult to make nice UIs with liblayout.

	// View hierarchy:
	//
	// MWindow
	// + MBorder
	//   + VGroup
	//     + HGroup
	//       + app icon view
	//       + VGroup
	//         + Space
	//         + MStringView - be_bold_font, size * 1.2
	//     + HGroup
	//       + Space
	//       + VGroup
	//         + MStringView - be_plain_font
	//         + MStringView - be_plain_font - right-justified
	//         + Space
	//         + HGroup
	//           + VGroup
	//             + MStringView - be_plain_font - right-justified
	//             + Space
	//             + MStringView - be_plain_font - right-justified
	//             + MStringView - be_plain_font - right-justified
	//			   + Space
	//             + MStringView - be_plain_font - right-justified
	//             + Space
	//           + VGroup
	//             + MStringView - be_plain_font - centre-justified
	//             + Space
	//             + MStringView - be_plain_font - centre-justified
	//             + MStringView - be_plain_font - centre-justified
	//             + Space
	//             + MStringView - be_plain_font - centre-justified
	//             + Space
	//           + VGroup
	//             + MStringView - be_bold_font - left-justified
	//             + Space
	//             + MStringView - be_bold_font - left-justified
	//             + MStringView - be_bold_font - left-justified
	//             + Space
	//             + MStringView - be_bold_font - left-justified
	//             + MStringView - be_bold_font - left-justified

	// TODO: This doesn't quite line up the way I wanted, especially by
	// 		 the bottom of the window...
	MBorder *view = new MBorder(
						M_RAISED_BORDER, 16, "",
						new VGroup(
							new HGroup(
								new MBViewWrapper( new AppIconView ),
								new Space( minimax( 8, 0, 8, 65536 ) ),

								new VGroup(
									new Space,
									( title = new MStringView( app_version_text ) ),
									NULL
								),
								NULL
							),
							
							new Space( minimax( 0, 16, 65536, 16 ) ),
							
							new HGroup(
								new Space( minimax( 32, 0, 32, 65536 ) ),
								new Space( minimax( 8, 0, 8, 65536 ) ),

								new VGroup(
									new MStringView( about_copyright, B_ALIGN_RIGHT ),
									( email = new URLStringView( about_email, "mailto:arcanedragon@pobox.com", B_ALIGN_RIGHT ) ),

									new Space( minimax( 0, 16, 65536, 16 ) ),
								
									new HGroup(
										new VGroup(
											new MStringView( about_design, B_ALIGN_RIGHT ),
											new Space( minimax( 0, 16, 65536, 16 ) ),
											new URLStringView( about_liblayout, about_liblayout_url, B_ALIGN_RIGHT ),
											new URLStringView( about_sgb, about_sgb_url, B_ALIGN_RIGHT ),
											new URLStringView( about_sfx, about_sfx_url, B_ALIGN_RIGHT ),
#if defined(SHAREWARE_NAG)
											new Space( minimax( 0, 16, 65536, 16 ) ),
											new MStringView( about_user, B_ALIGN_RIGHT ),
											new Space( minimax( 0, 16, 65536, 16 ) ),
#endif
											NULL
										),
										new Space( minimax( 8, 0, 8, 65536 ) ),
										
										new VGroup(
											new MStringView( about_emdash, B_ALIGN_CENTER ),
											new Space( minimax( 0, 16, 65536, 16 ) ),
											new MStringView( about_emdash, B_ALIGN_CENTER ),
											new MStringView( about_emdash, B_ALIGN_CENTER ),
											new MStringView( about_emdash, B_ALIGN_CENTER ),
#if defined(SHAREWARE_NAG)
											new Space( minimax( 0, 16, 65536, 16 ) ),
											new MStringView( about_emdash, B_ALIGN_CENTER ),
											new Space( minimax( 0, 16, 65536, 16 ) ),
#endif
											NULL
										),
										new Space( minimax( 8, 0, 8, 65536 ) ),
	
										new VGroup(
											( chrish = new MStringView( about_designer ) ),
											new Space( minimax( 0, 16, 65536, 16 ) ),
											( marcone = new MStringView( about_liblayout_author ) ),
											( bteitz = new MStringView( about_sgb_author ) ),
											( pir = new MStringView( about_sfx_author ) ),
#if defined(SHAREWARE_NAG)
											new Space( minimax( 0, 16, 65536, 16 ) ),
											( reg_name = new MStringView( about_reg_name ) ),
											new MStringView( about_reg_company ),
#endif
											NULL
										),

										NULL
									),

									NULL
								),
							
								NULL
							),

							NULL
						)
					);

	AddChild( view );

	BFont title_font( be_bold_font );
	title_font.SetSize( title_font.Size() * 1.2 );
	title->SetFont( &title_font );

	email->SetFont( be_fixed_font );
	
	chrish->SetFont( be_bold_font );
	marcone->SetFont( be_bold_font );
	bteitz->SetFont( be_bold_font );
	pir->SetFont( be_bold_font );
#if defined(SHAREWARE_NAG)
	reg_name->SetFont( be_bold_font );
#endif

	BPoint pos;
	if( settings->FindPoint( SETTING_POSITION, &pos ) == B_OK ) {
		if( pos.x != 0.0 && pos.y != 0.0 ) MoveTo( pos );
	}

	Show();
}

AboutWindow::~AboutWindow( void )
{
	if( _sfx ) delete _sfx;
}

// ----------------------------------------------------------------------
// BWindow methods

void AboutWindow::MessageReceived( BMessage *msg )
{
	switch( msg->what ) {
	case APP_ICON_CLICK:
		{
			BResources *rsrc = be_app->AppResources();
			const void *sfx_data = rsrc->LoadResource( B_RAW_TYPE, "liquid pouring", &_size );
			BPath sfx_path( "/tmp/StartUp!_liquid_pouring" );
			BEntry sfx_ent( sfx_path.Path() );
			BFile *sfx_file = new BFile( &sfx_ent, B_WRITE_ONLY | B_CREATE_FILE );
			sfx_file->Write( sfx_data, _size );
			delete sfx_file;
		
			_sfx = new BFileGameSound( sfx_path.Path(), false );
			if( _sfx->InitCheck() != B_OK ) {
				// ugh...
				delete _sfx;
				_sfx = NULL;
			} else {
				_sfx->Preload();
				_sfx->StartPlaying();
			}
		
			sfx_ent.Remove();
		}
		break;

	default:
		BWindow::MessageReceived( msg );
		break;
	}
}

bool AboutWindow::QuitRequested( void )
{
	App *app = static_cast<App *>( be_app );
	StartupSettings *settings = app->GetSettings();
	settings->SavePoint( SETTING_POSITION, Frame().LeftTop() );

	app->DeregisterWindow( this );

	if( _sfx && _sfx->IsPlaying() ) _sfx->StopPlaying();

	return true;
}

// ----------------------------------------------------------------------
// Private methods

const char *AboutWindow::VersionVariety( uint32 variety )
{
	BResources *rsrc = be_app->AppResources();

	switch( variety ) {
	case 0:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:development", &_size ) );
		break;
	case 1:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:alpha", &_size ) );
		break;
	case 2:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:beta", &_size ) );
		break;
	case 3:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:gamma", &_size ) );
		break;
	case 4:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:gold", &_size ) );
		break;
	case 5:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:final", &_size ) );
		break;
	default:
		return static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "variety:unknown", &_size ) );
		break;
	}
	
	// You won't get here.
	return NULL;
}
