// ----------------------------------------------------------------------
// StartUp! - Application object
//
// Copyright © 2000 Arcane Dragon Software (arcanedragon@pobox.com)
//
// Handle application duties for StartUp!; keeps track of our StartupItems,
// resources, etc.
// ----------------------------------------------------------------------

#include "App.h"
#include "AboutWindow.h"
#include "BufferedFile.h"
#include "CommandParser.h"
#include "LaunchWindow.h"
#include "MainWindow.h"
#include "MD5.h"
#include "MutexList.h"
#include "NagWindow.h"
#include "RegistrationWindow.h"
#include "StartupItem.h"
#include "StartupSettings.h"
#include "SysLog.h"

#include <app/Message.h>
#include <app/Roster.h>
#include <interface/Alert.h>
#include <kernel/fs_attr.h>
#include <kernel/OS.h>
#include <storage/AppFileInfo.h>
#include <storage/Directory.h>
#include <storage/Entry.h>
#include <storage/File.h>
#include <storage/FindDirectory.h>
#include <storage/Mime.h>
#include <storage/Path.h>
#include <storage/Resources.h>
#include <storage/StorageDefs.h>
#include <support/syslog.h>
#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// ----------------------------------------------------------------------
// Constructor/destructor
App::App( void )
	: BApplication( static_cast<const char *>( AppResources()->LoadResource( B_MIME_STRING_TYPE, "BEOS:APP_SIG", &_app_sig_size ) ) ),
	  _exit_value( EXIT_SUCCESS ),
	  _info( NULL ),
	  _app_file( NULL ),
	  _startup_path( NULL ),
	  _startup_items( NULL ),
	  _do_launch( false ),
	  _do_install( false ),
	  _windows( NULL ),
	  _settings( NULL )
{
	_windows = new MutexList( "window list" );

	SysLog_Open( "syslog:name" );

	// Find and load our settings
	_settings = new StartupSettings;
	if( _settings->InitCheck() != B_OK ) {
		// that's bad...
		PostMessage( B_QUIT_REQUESTED );
		return;
	}

	status_t retval = _settings->Load();
	if( retval != B_OK ) {
		// we're using defaults
	}

	// Check to see if we've got a place to run from.  If not, we want to
	// bail early.
	if( !CheckStartupPath() ) {
		// That's bad, we can't run.
		size_t size = 0;
		BResources *rsrc = AppResources();

		const char *bail_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "bail title", &size ) );
		const char *bail_text = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "bail alert text", &size ) );
		const char *bail_button = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "bail button", &size ) );
		
		BAlert *bail = new BAlert( bail_title,
								   bail_text,
								   bail_button, NULL, NULL,
								   B_WIDTH_AS_USUAL,
								   B_IDEA_ALERT );
		int32 which = bail->Go();
		
		// Shut up gcc...
		which = 0;

		SysLog_Info( "syslog:bail" );
		
		PostMessage( B_QUIT_REQUESTED );
	}
}

App::~App( void )
{
	// Save our settings
	if( _settings ) {
		status_t retval = _settings->Save();
		if( retval != B_OK ) {
			// dang
		}
		delete _settings;
	}

	if( _startup_items ) {
		if( _startup_items->Lock() ) {
			for( int32 i = _startup_items->CountItems(); --i >= 0; ) {
				StartupItem *item = static_cast<StartupItem *>( _startup_items->RemoveItem( i ) );
				if( item ) delete item;
			}
		}

		delete _startup_items;
	}

	if( _info ) delete _info;
	if( _app_file ) delete _app_file;
	if( _startup_path ) delete _startup_path;
	if( _windows ) delete _windows;

	SysLog_Close();
}

// ----------------------------------------------------------------------
// BApplication hooks

void App::AboutRequested( void )
{
	BAppFileInfo *info = GetInfo();
	if( !info ) {
		// Serious problem here!  If we can't get this, we should probably
		// bail and run screaming for the hills...
		size_t size = 0;
		BResources *rsrc = AppResources();

		const char *about_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about title", &size ) );
		const char *about_text = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about alert text", &size ) );
		const char *about_button = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "about button", &size ) );
		
		BAlert *about = new BAlert( about_title,
									about_text,
									about_button, NULL, NULL,
									B_WIDTH_AS_USUAL,
									B_IDEA_ALERT );
		int32 which = about->Go();
		
		// Shut up gcc...
		which = 0;
	} else {
		// Display the full-on About dialog.
		AboutWindow *about = new AboutWindow;
		if( about == NULL ) {
			SysLog_Error( "syslog:no about" );
			return;
		}
	}
}

void App::ArgvReceived( int32 argc, char **argv )
{
	// Handle command-line arguments
	size_t size = 0;
	BResources *rsrc = AppResources();
	const char *launch_arg = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "arg:launch", &size ) );
	const char *install_arg = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "arg:install", &size ) );

	for( int i = 1; i < argc; i++ ) {
		if( !strcasecmp( argv[i], launch_arg ) ) {
			_do_launch = true;
		} else if( !strcasecmp( argv[i], install_arg ) ) {
			_do_install = true;
		} else {
			SysLog_Error( "syslog:unknown arg", argv[i] );
		}
	}

	// Launching and installing are two separate things.
	if( _do_launch && _do_install ) _do_launch = false;
}

void App::MessageReceived( BMessage *msg )
{
	switch( msg->what ) {
	case NAG_SHOW_WINDOW:
#if defined(SHAREWARE_NAG)
		{
			NagWindow *win = new NagWindow;
			win = win;	// shut up gcc...
		}
#endif
		break;

	case NAG_ENTER_KEY:
#if defined(SHAREWARE_NAG)
		{
			RegistrationWindow *win = new RegistrationWindow;
			win = win;	// shut up gcc...
		}
#endif
		break;
	
	default:
		BApplication::MessageReceived( msg );
		break;
	}
}

bool App::QuitRequested( void )
{
	// We'll need some usefulness here to handle the case where we've got
	// some open windows...
#if defined(SHAREWARE_NAG)
	if( _is_nagging ) return false;
#endif

	if( _windows->Lock() ) {
		if( _windows->CountItems() == 0 ) return true;
		for( int32 i = 0; i < _windows->CountItems(); i++ ) {
			BWindow *win = static_cast<BWindow *>( _windows->ItemAt( i ) );
			win->PostMessage( B_QUIT_REQUESTED );
		}
		_windows->Unlock();
	}

	return false;
}

void App::ReadyToRun( void )
{
	// Do something useful
	BWindow *a_window = NULL;

#if defined(SHAREWARE_NAG)
	if( !CheckKey() ) {
		// if not registered, nag...
		SetIsNagging( true );

		a_window = new NagWindow;
	}
#endif

	// If there are no items, and we're not doing an install, we can bail
	// out early.
	MutexList *items = GetStartupItems();
	if( items && items->CountItems() == 0 && !_do_install ) {
		PostMessage( B_QUIT_REQUESTED );
		return;
	}

	if( _do_launch ) {
		// do the launch
		a_window = new LaunchWindow( items );
	} else if( _do_install ) {
		// do the install
		DoInstall();
	} else {
		// display the interface
		a_window = new MainWindow( items );
	}

	// This is to shut up gcc... we'll always have a window.
	if( !a_window ) PostMessage( B_QUIT_REQUESTED );
}

void App::RefsReceived( BMessage *msg )
{
	// Someone has dropped items on us; add them to the Startup Items
	// directory.
	type_code type = 0;
	int32 num = 0;
	status_t retval = msg->GetInfo( "refs", &type, &num );
	if( num < 1 ) {
		// Someone's playing tricks on us...
		return;
	}

	for( int32 i = 0; i < num; i++ ) {
		// For each ref, we should create an item.  We'll have to make up
		// a new name if the item's name already exists there.
		entry_ref ref;
		retval = msg->FindRef( "refs", i, &ref );
		if( retval != B_OK ) {
			// Couldn't get that ref, try the next...
			continue;
		}

		// Find this item's name...
		BPath ref_path( &ref );
		if( ref_path.InitCheck() != B_OK ) {
			// Couldn't create a path for that entry...
			continue;
		}

		if( !CreateItem( ref_path.Leaf(), ref_path.Path() ) ) {
			// waah!
		}
	}
}

// ----------------------------------------------------------------------
// StartUp application methods

int App::ExitValue( void )
{
	return _exit_value;
}

void App::SetExitValue( int val )
{
	_exit_value = val;
}

BAppFileInfo *App::GetInfo( void )
{
	// Lazy-allocate the application's BAppFileInfo.
	if( _info ) return _info;

	BFile *file = GetAppFile();
	if( file == NULL ) {
		SysLog_Error( "syslog:no app file" );
		return NULL;
	}

	_info = new BAppFileInfo( file );
	if( _info->InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no app file info", _info->InitCheck() );

		delete _info;
		_info = NULL;
	}
	
	return _info;
}

BFile *App::GetAppFile( void )
{
	// Lazy-allocate the application's BFile; needed to get its BAppFileInfo,
	// which is the useful bit we want.
	if( _app_file ) return _app_file;

	// app_info will get the application file's entry_ref, 
	// which we can turn into a BFile.
	app_info appi;
	status_t retval = GetAppInfo( &appi );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no app info", retval );

		return NULL;
	}
	
	_app_file = new BFile( &appi.ref, B_READ_ONLY );
	if( _app_file->InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no app file", _app_file->InitCheck() );

		delete _app_file;
		_app_file = NULL;
	}

	return _app_file;
}

BPath *App::GetStartupPath( void )
{
	// Lazy-allocate the StartUp items folder, inside the B_USER_DIRECTORY 
	// directory.
	if( _startup_path ) return _startup_path;

	_startup_path = new BPath;

	status_t retval = find_directory( B_USER_DIRECTORY, _startup_path, true );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no user dir", retval );

		delete _startup_path;
		_startup_path = NULL;
	} else {
		BResources *rsrc = AppResources();
		size_t size;
		const char *startup_folder = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "StartUp! folder", &size ) );
		retval = _startup_path->Append( startup_folder, true );
		if( retval != B_OK ) {
			SysLog_Error( "syslog:no startup dir", retval );

			delete _startup_path;
			_startup_path = NULL;
		}
	}

	return _startup_path;
}

MutexList *App::GetStartupItems( void )
{
	// Lazy allocate the Startup Items list, and populate it.
	if( _startup_items ) return _startup_items;

	// We'll need the path before we can look through the items...
	BPath *path = GetStartupPath();
	if( path == NULL ) {
		SysLog_Error( "syslog:no startup path" );

		return NULL;
	}

	BDirectory item_dir( path->Path() );
	if( item_dir.InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no startup dir", item_dir.InitCheck() );
		
		return NULL;
	}

	// Create the new list of items
	_startup_items = new MutexList( "Startup Items list" );

	// Read the entries and make StartupItems out of them.
	_startup_items->Lock();

	BEntry ent;
	while( item_dir.GetNextEntry( &ent ) == B_OK ) {
		StartupItem *item = new StartupItem( &ent );
		if( item->InitCheck() == B_OK ) {
			if( !_startup_items->AddItem( item ) ) {
				// d'oh!
				SysLog_Error( "syslog:add item failed" );
			}
		} else {
			// problems with that item...
			SysLog_Error( "syslog:item init failed", item->InitCheck() );
		}
	}

	_startup_items->Unlock();
	
	return _startup_items;
}

bool App::ShowDocs( void )
{
	// Display the documentation, which is an HTML file in the application's
	// directory.
	BResources *rsrc = AppResources();
	size_t size = 0;
	const char *doc_file = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "docs file", &size ) );

	// Find the documentation, safely stashed in the application's directory.
	//
	// We go through the entry_ref in app_info again because you can't go
	// through a BFile.  Strange but true.  Also note how you can't create
	// a BPath from an entry_ref...
	app_info appi;
	status_t retval = GetAppInfo( &appi );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no app info", retval );

		return false;
	}

	BPath path( &appi.ref );
	if( path.InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no app path", path.InitCheck() );

		return false;
	}

	retval = path.GetParent( &path );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no app parent path", retval );

		return false;
	}
	retval = path.Append( doc_file );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no doc path", retval );

		return false;
	}

	// Now that we've got the path to the docs, we need to have an entry_ref
	// to give to the Roster.  Sheesh.
	entry_ref doc_ref;
	retval = get_ref_for_path( path.Path(), &doc_ref );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no doc ref", retval );

		return false;
	}
	
	retval = be_roster->Launch( &doc_ref );
	if( retval != B_ALREADY_RUNNING && retval != B_OK ) {
		SysLog_Error( "syslog:no doc launch", retval );

		return false;
	}

	// Then it worked, yay!
	return true;
}

bool App::ShowADSWebsite( void )
{
	// Display the Arcane Dragon Software website in the URL handler.
	BResources *rsrc = AppResources();
	size_t size = 0;
	const char *ads_url = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "ADS url", &size ) );
	const char *url_handler = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "URL handler MIME", &size ) );

	// Find the URL handler, and tell it to go to our URL.  Hopefully.
	int argc = 1;
	char *argv[1];
	argv[0] = const_cast<char *>( ads_url );

	status_t retval = be_roster->Launch( url_handler, argc, argv );
	if( retval != B_ALREADY_RUNNING && retval != B_OK ) {
		SysLog_Error( "syslog:no url launch", retval );

		return false;
	}

	// Then it worked, yay!
	return true;
}

bool App::CheckStartupPath( void )
{
	// Check to see if the startup path is there; this just checks whether it
	// exists or can be created, it doesn't look for StartUp items inside.
	BPath *startup_path = GetStartupPath();
	if( startup_path == NULL ) {
		SysLog_Error( "syslog:no startup path" );

		return false;
	}
	
	BPath parent_path;
	status_t retval = startup_path->GetParent( &parent_path );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no startup parent path", retval );

		return false;
	}
	
	BDirectory parent_dir( parent_path.Path() );
	retval = parent_dir.InitCheck();
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no startup parent dir", retval );

		return false;
	}
	
	// If the StartUp items directory isn't there, try to create it.
	if( !parent_dir.Contains( startup_path->Leaf(), B_DIRECTORY_NODE ) ) {
		retval = parent_dir.CreateDirectory( startup_path->Leaf(), NULL );
		if( retval != B_OK ) {
			SysLog_Error( "syslog:no create startup dir", retval );

			return false;
		}
	}

	// If you got here, the directory exists.  Yay!
	return true;
}

// ----------------------------------------------------------------------
// Convert a UserBootscript into multiple Startup Items

status_t App::ConvertScript( directory_which where, const char *name )
{
	// The idea is to go through the given script (name in where), and do
	// these things:
	//
	// - save the original name as name.pre-StartUp
	//
	// - preserve any lines we don't understand; put lines like:
	//		# StartUp! couldn't convert this ---- START ----
	//		# StartUp! couldn't convert this ---- END ----
	//	 around the items we don't grok
	//
	// - put a comment block at the top stating that StartUp! modified this
	//
	// - convert any lines we do understand into Startup Items with args
	//	 (if necessary); preserve the original line as a comment
	//
	// - add StartUp itself before the first command we did understand

	// Strings we'll need.
	BResources *rsrc = AppResources();
	const char *backup_ext = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "backup extension", &_app_sig_size ) );
	const char *edited_by = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "edited by StartUp", &_app_sig_size ) );
	const char *startup_comment = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "StartUp comment", &_app_sig_size ) );
	const char *launch_arg = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "arg:launch", &_app_sig_size ) );
	const char *shell_magic = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "shell magic", &_app_sig_size ) );
	const char *cmd_prefix1 = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "command prefix 1", &_app_sig_size ) );
	const char *cmd_prefix2 = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "command prefix 2", &_app_sig_size ) );

	// Find the file we want to convert.
	BPath orig_path;
	status_t retval = find_directory( where, &orig_path, false );
	if( retval != B_OK ) return retval;

	BDirectory dir( orig_path.Path() );
	if( dir.InitCheck() != B_OK ) return dir.InitCheck();

	BufferedFile orig_file( &dir, name, B_READ_ONLY );
	if( orig_file.InitCheck() != B_OK ) return orig_file.InitCheck();

	// Now we've got the file open; we should rename it with the backup
	// extension, and create the new file.
	BEntry ent;
	retval = dir.FindEntry( name, &ent, false );
	if( retval != B_OK ) return retval;

	BString new_name( name );
	new_name.Append( backup_ext );
	retval = ent.Rename( new_name.String(), true );
	if( retval != B_OK ) {
		retval = ent.Remove();
		if( retval != B_OK ) return retval;
	}

	BufferedFile new_file( &dir, name, B_WRITE_ONLY | B_CREATE_FILE | B_ERASE_FILE );
	if( new_file.InitCheck() != B_OK ) return new_file.InitCheck();

	// Copy attributes from the old file to the new file
	{
		BPath full_path( orig_path );
		full_path.Append( new_name.String() );
		BNode orig_node( full_path.Path() );
		full_path.SetTo( orig_path.Path() );
		full_path.Append( name );
		BNode new_node( full_path.Path() );

		if( orig_node.InitCheck() != B_OK ) {
			SysLog_Error( "syslog:no item node", orig_node.InitCheck() );
		}
		if( new_node.InitCheck() != B_OK ) {
			SysLog_Error( "syslog:no item node", new_node.InitCheck() );
		}

		char attr_name[B_ATTR_NAME_LENGTH];
		while( orig_node.GetNextAttrName( attr_name ) == B_OK ) {
			attr_info info;
			if( ( retval = orig_node.GetAttrInfo( attr_name, &info ) ) != B_OK ) {
				SysLog_Error( "syslog:no attr info", retval );
				continue;
			}

			char *buff = new char[info.size];
			if( orig_node.ReadAttr( attr_name, info.type, 0, buff, info.size ) != info.size ) {
				SysLog_Error( "syslog:no attr read", attr_name );
				delete [] buff;
				continue;
			}

			if( new_node.WriteAttr( attr_name, info.type, 0, buff, info.size ) != info.size ) {
				SysLog_Error( "syslog:no attr write", attr_name );
				delete [] buff;
				continue;
			}

			delete [] buff;
		}
	}

	const char *slash_n = "\n";
	new_file.BufferedWrite( shell_magic, strlen( shell_magic ) );
	new_file.BufferedWrite( slash_n, strlen( slash_n ) );
	new_file.BufferedWrite( edited_by, strlen( edited_by ) );

	int32 lines_processed = 0;		// # of lines, so we can catch #! comment, etc.
	char buff[ARG_MAX];				// line buffer
	CommandParser parser;			// partial bash parser
	BMessage cmd;					// command and args
	BMessage lines;					// store lines when we're not done yet
	bool added_startup = false;		// Have we added StartUp! itself yet?
	while( orig_file.BufferedReadLine( buff, ARG_MAX ) > 0 ) {
		lines_processed++;
		switch( parser.Parse( buff, cmd ) ) {
		case B_OK:
			// Got a command

			// If we haven't already put StartUp! in there, we should add
			// it now.
			if( !added_startup ) {
				added_startup = true;

				char comment[ARG_MAX];
				snprintf( comment, ARG_MAX, startup_comment, name );

				app_info appi;
				retval = GetAppInfo( &appi );
				if( retval != B_OK ) {
					//SysLog_Error( "syslog:no app info", retval );
			
					//return NULL;
				}
				ent.SetTo( &appi.ref );
				BPath path( &ent );
				
				new_file.BufferedWrite( slash_n, strlen( slash_n ) );
				new_file.BufferedWrite( comment, strlen( comment ) );
				new_file.BufferedWrite( slash_n, strlen( slash_n ) );
				new_file.BufferedWrite( path.Path(), strlen( path.Path() ) );
				new_file.BufferedWrite( " ", 1 );
				new_file.BufferedWrite( launch_arg, strlen( launch_arg ) );
				new_file.BufferedWrite( slash_n, strlen( slash_n ) );
			}
			
			// Write out anything that we've processed before.
			{
				type_code type;
				int32 count;
				lines.GetInfo( "lines", &type, &count );
				if( type == B_STRING_TYPE && count > 0 ) {
					const char *str = NULL;
					int32 idx = 0;
					while( lines.FindString( "lines", idx++, &str ) == B_OK ) {
						new_file.BufferedWrite( str, strlen( str ) );
						new_file.BufferedWrite( slash_n, strlen( slash_n ) );
					}
				}
				
				lines.MakeEmpty();
			}

			// Write out the command we recognized as a comment.
			if( !cmd.HasBool( "complex" ) ) {
				new_file.BufferedWrite( cmd_prefix1, strlen( cmd_prefix1 ) );
				BPath *item_dir = GetStartupPath();
				new_file.BufferedWrite( item_dir->Leaf(), strlen( item_dir->Leaf() ) );
				new_file.BufferedWrite( cmd_prefix2, strlen( cmd_prefix2 ) );
				new_file.BufferedWrite( buff, strlen( buff ) );
				new_file.BufferedWrite( slash_n, strlen( slash_n ) );

				// Create the new Startup Item.
				retval = NewItem( cmd );
				if( retval != B_OK ) {
					// whine!
				}
			} else {
				new_file.BufferedWrite( buff, strlen( buff ) );
				new_file.BufferedWrite( slash_n, strlen( slash_n ) );
			}
			cmd.MakeEmpty();
			break;
			
		case B_ERROR:
			// Something bad
			lines.AddString( "lines", buff );
			cmd.MakeEmpty();
			break;
			
		case B_BUSY:
			// Continued line
			lines.AddString( "lines", buff );
			break;
			
		case B_CANCELED:
			// Comment or blank; skip the magic comment at the start
			if( lines_processed > 1 ) lines.AddString( "lines", buff );
			cmd.MakeEmpty();
			break;
			
		default:
			break;
		}
	}

	// If there's anything left in the lines message, it's all unprocessed.
	type_code type;
	int32 count;
	lines.GetInfo( "lines", &type, &count );
	if( type == B_STRING_TYPE && count > 0 ) {
		const char *str = NULL;
		int32 idx = 0;
		while( lines.FindString( "lines", idx++, &str ) == B_OK ) {
			new_file.BufferedWrite( str, strlen( str ) );
			new_file.BufferedWrite( slash_n, strlen( slash_n ) );
		}

		lines.MakeEmpty();
	}

	// Ok, this is bloody stupid.
	new_file.FlushBuffer();

	// Offer to fire up the new UserBootscript in the user's favourite
	// editor?

	return B_OK;
}

bool App::CreateItem( const char *name, const char *src, BEntry *an_ent )
{
	if( name == NULL || src == NULL ) return false;
	if( strlen( name ) == 0 || strlen( src ) == 0 ) return false;

	// Create a new item with the given name in the Startup Items directory,
	// adjusting it as necessary.
	BPath *base_path = GetStartupPath();
	BDirectory base_dir( base_path->Path() );

	if( base_dir.Contains( name ) ) {
		BResources *rsrc = AppResources();
		size_t size = 0;
		const char *fmt = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "new item format", &size ) );

		char buff[B_FILE_NAME_LENGTH];
		int x = 1;
		do {
			snprintf( buff, B_FILE_NAME_LENGTH, fmt, name, x );
			x++;
		} while( base_dir.Contains( buff ) );

		status_t retval = base_dir.CreateSymLink( buff, src, NULL );
		if( retval != B_OK ) {
			// Error creating new symlink
			return false;
		}

		BPath p( *base_path );
		p.Append( buff );
		if( an_ent ) an_ent->SetTo( p.Path(), false );
	} else {
		status_t retval = base_dir.CreateSymLink( name, src, NULL );
		if( retval != B_OK ) {
			// Error creating new symlink
			return false;
		}

		BPath p( *base_path );
		p.Append( name );
		if( an_ent ) an_ent->SetTo( p.Path(), false );
	}

	return true;
}

status_t App::NewItem( BMessage& msg )
{
	const char *cmd = NULL;
	if( msg.FindString( "cmd", &cmd ) != B_OK ) return B_CANCELED;

	// If the cmd has ~/ at the front, we should expand that to a full
	// path.
	BString the_cmd( cmd );
	if( the_cmd[0] == '~' && the_cmd[1] == '/' ) {
		the_cmd.RemoveFirst( "~" );
		BPath user_dir;
		if( find_directory( B_USER_DIRECTORY, &user_dir ) != B_OK ) {
			const char *home = getenv( "HOME" );
			if( home ) {
				BString new_cmd( home );
				new_cmd.Append( the_cmd );
				the_cmd.SetTo( new_cmd.String() );
			}
		} else {
			BString new_cmd( user_dir.Path() );
			new_cmd.Append( the_cmd );
			the_cmd.SetTo( new_cmd );
		}
	}

	// If the cmd isn't an absolute path, we'll have to pick one by
	// searching the PATH.
	if( the_cmd[0] != '/' ) {
		bool found = false;

		char *tmp = strdup( getenv( "PATH" ) );
		char *path = tmp;
		char *part = tmp;

		while( part ) {
			part = strtok_r( path, ":", &path );
			BPath p( part );
			p.Append( the_cmd.String() );
			BEntry ent( p.Path(), false );
			if( ent.Exists() ) {
				the_cmd.SetTo( p.Path() );
				found = true;
				break;
			}
		}

		free( tmp );

		if( !found ) return B_FILE_NOT_FOUND;
	}

	// Create the symlink for the new item
	BEntry new_item;
	BPath the_path( the_cmd.String(), NULL, true );
	if( !CreateItem( the_path.Leaf(), the_path.Path(), &new_item ) ) return B_ENTRY_NOT_FOUND;

	StartupItem the_item( &new_item );
	if( !the_item.SetPrefs( msg ) ) return B_ERROR;

	return B_OK;
}

// ----------------------------------------------------------------------
// Window registry, so we know when/how to exit

void App::RegisterWindow( BWindow *win )
{
	// We've opened a new window...
	if( _windows->Lock() ) {
		_windows->AddItem( (void *)win );
		_windows->Unlock();
	}
}

void App::DeregisterWindow( BWindow *win )
{
	// We've closed a window... if that's it, we should quit.
	if( _windows->Lock() ) {
		_windows->RemoveItem( (void *)win );
		if( _windows->CountItems() == 0 ) PostMessage( B_QUIT_REQUESTED );
		_windows->Unlock();
	}
}

// ----------------------------------------------------------------------
// Settings methods

StartupSettings *App::GetSettings( void )
{
	return _settings;
}

// ----------------------------------------------------------------------
// Registration key methods
//
// This code is uglier than it could be because I'm using the md5.c from
// RSA (via the Python distribution, actually).  Casts-r-us.  If I wasn't
// feeling lazy, I'd rewrite it to operate with BStrings instead... oh
// well it works.
//
// The key generation algorithm comes up with a 6-byte hex number
// in the form:
//
// 00-0000-0000-00
//
// This is 6 bytes worth of data (eg. 12-3456-789a-bc), generated from the 
// input via:
//
// md5 digest of company, then name == 128 bits (16 bytes)
// add the string version of copies to the digest if atol() fails on it
//
// Mutated with #copies into 6 bytes via:
//
// key[0] = md5[7] ^ md5[10] ^ copies[0]
// key[1] = md5[5] ^ md5[11] ^ md5[15]
// key[2] = md5[0] ^ md5[1] ^ md5[9] ^ md5[13] ^ md5[14]
// key[3] = md5[12]
// key[4] = md5[2] ^ md5[3] ^ md5[6] ^ copies[2] ^ copies[3]
// key[5] = md5[4] ^ md5[8] ^ copies[4]
//
// We don't store the key though, that's too easy to swipe, or something.
// Instead, we store an int32 key, generated thusly:
//
// code = key[0] << 8
// code += key[1]
// code += key[2] << 24
// code += key[3] << 16
// code ^= key[4] << 8
// code ^= key[5] << 25
//
// At runtime, we get the name, company, # of copies from prefs, generate
// a key, and compare that key's code vs. the code stored in the prefs.
// If they match, voila.
//
// TODO: Adding an MD5 digest attribute to the settings file to detect
//       tampering might make this more secure.  Depends how much effort
//       you want to put into it.

static inline int32 calc_code( unsigned char *key )
{
	int32 calc = 0;
	calc += ( key[0] << 8 );
	calc += ( key[1] );
	calc += ( key[2] << 24 );
	calc += ( key[3] << 16 );
	calc ^= ( key[4] << 8 );
	calc ^= ( key[5] << 24 );
	return calc;
}

static inline void calc_key( const char *digest, unsigned char *key, int32 copies )
{
	key[0] = digest[7] ^ digest[10] ^ (unsigned char)((copies & 0xff000000) >> 24);
	key[1] = digest[5] ^ digest[11] ^ digest[15];
	key[2] = digest[0] ^ digest[1] ^ digest[9] ^ digest[13] ^ digest[14];
	key[3] = digest[12];
	key[4] = digest[2] ^ digest[3] ^ digest[6] ^ (unsigned char)((copies & 0x00ff0000) >> 16) ^ (unsigned char)((copies & 0x0000ff00) >> 8);
	key[5] = digest[4] ^ digest[8] ^ (unsigned char)(copies & 0x000000ff);
}

bool App::CheckKey( void )
{
#if !defined(SHAREWARE_NAG)
	return true;
#else
	const char *name = NULL;
	const char *company = NULL;
	const char *copies_str = NULL;
	int32 code = 0;

	StartupSettings *settings = GetSettings();
	if( settings->FindString( "reg:name", &name ) != B_OK ) return false;
	if( settings->FindString( "reg:company", &company ) != B_OK ) return false;
	if( settings->FindString( "reg:copies", &copies_str ) != B_OK ) return false;
	if( settings->FindInt32( "reg:code", &code ) != B_OK ) return false;

	int32 copies = atol( copies_str );

	MD5 digester;
	digester.Update( (unsigned char *)const_cast<char *>( company ) );
	digester.Update( (unsigned char *)const_cast<char *>( name ) );
	if( copies == 0 ) digester.Update( (unsigned char *)const_cast<char *>( copies_str ) );
	BString *digest_string = digester.Digest();

	unsigned char key[6];
	calc_key( digest_string->String(), key, copies );

	delete digest_string;

	int32 calc = calc_code( key );

	if( calc != code ) return false;

	// Time bomb certain keys.
	if( !strncasecmp( name, "beta", 4 ) ) {
		// Find the target value thusly:
		//
		// import time
		// time.mktime( ( 2000, 8, 20, 12, 0, 0, 0, 0, 0 ) )
		//
		// beta 1 - 966787200 (August 20/2000, 12:00 noon)
		if( real_time_clock() >= 966787200 ) return false;
	}

	return true;
#endif
}

static inline int hex2int( const char c )
{
	if( c >= '0' && c <= '9' ) {
		return c - '0';
	} else if( c >= 'a' && c <= 'f' ) {
		return c - 'a' + 0x0a;
	} else if( c >= 'A' && c <= 'F' ) {
		return c - 'A' + 0x0a;
	}
	
	return 0;
}

bool App::TestKey( BString& a_name, BString& a_company, BString &a_copies, 
				   BString &a_key )
{
#if !defined(SHAREWARE_NAG)
	return true;
#else
	// Record the values we're testing with.
	SysLog_Info( "syslog:test name", a_name.String() );
	SysLog_Info( "syslog:test company", a_company.String() );
	SysLog_Info( "syslog:test copies", a_copies.String() );
	SysLog_Info( "syslog:test key", a_key.String() );

	// Convert the string key into a 6-byte key.
	unsigned char real_key[6];
	memset( real_key, 0, sizeof( unsigned char ) * 6 );
	int idx = 0;
	int val = 0;
	int digits = 0;
	for( int i = 0; i < a_key.Length() && idx < 6; i++ ) {
		if( isxdigit( a_key[i] ) ) {
			val <<= 4;
			val = val + hex2int( a_key[i] );
			digits++;
		}
		
		if( digits == 2 ) {
			real_key[idx++] = (unsigned char)val;
			digits = 0;
			val = 0;
		}
	}

	char *name = a_name.LockBuffer( 0 );
	char *company = a_company.LockBuffer( 0 );
	char *copies = a_copies.LockBuffer( 0 );

	// Add all the relevent bits to the MD5 hash.
	MD5 digester;
	digester.Update( (unsigned char *)company );
	digester.Update( (unsigned char *)name );
	int32 num_copies = atol( copies );
	if( num_copies == 0 ) digester.Update( (unsigned char *)copies );

	a_name.UnlockBuffer();
	a_company.UnlockBuffer();
	a_copies.UnlockBuffer();

	BString *digest_string = digester.Digest();

	// Mutate digest plus number of copies into 6-byte code.
	unsigned char key[6];
	calc_key( digest_string->String(), key, num_copies );

	delete digest_string;

	if( strncmp( (const char *)key, (const char *)real_key, 6 ) ) return false;

	// Calculate the magic code for this key
	int32 calc = calc_code( key );

	// Write this info to the settings
	StartupSettings *settings = GetSettings();

	settings->RemoveName( "reg:code" );
	settings->AddInt32( "reg:code", calc );

	settings->RemoveName( "reg:name" );
	settings->AddString( "reg:name", a_name.String() );
	settings->RemoveName( "reg:company" );
	settings->AddString( "reg:company", a_company.String() );
	settings->RemoveName( "reg:copies" );
	settings->AddString( "reg:copies", a_copies.String() );

	SetIsNagging( false );

	return true;
#endif
}

void App::SetIsNagging( bool val )
{
	_is_nagging = val;
}

// ----------------------------------------------------------------------
// Installation methods

void App::DoInstall( void )
{
	// Make sure they really want to do this.
	size_t size = 0;
	BResources *rsrc = AppResources();

	const char *install_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "install title", &size ) );
	const char *install_text = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "install alert text", &size ) );
	const char *install_cancel = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "install cancel button", &size ) );
	const char *install_do_it = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "install do it button", &size ) );
	
	BAlert *install = new BAlert( install_title,
								  install_text,
								  install_cancel, install_do_it, NULL,
								  B_WIDTH_AS_USUAL,
								  B_OFFSET_SPACING,
								  B_STOP_ALERT );
	int32 which = install->Go();

	// If they don't want to go through with it, bail.
	if( which == 0 ) {
		SetExitValue( B_OK );
		return;
	}

//	InstallWindow *a_window = new InstallWindow;	?
	SetExitValue( static_cast<int>( ConvertScript() ) );

	const char *install_done_text = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "install alert done text", &size ) );
	const char *install_done_button = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "install done button", &size ) );
	
	install = new BAlert( install_title,
						  install_done_text,
						  install_done_button, NULL, NULL,
						  B_WIDTH_AS_USUAL,
						  B_STOP_ALERT );
	which = install->Go();

	PostMessage( B_QUIT_REQUESTED );
}
