// ----------------------------------------------------------------------
// CommandParser - command-line parser
//
// Copyright © 2000 Arcane Dragon Software (arcanedragon@pobox.com)
//
// Simple Bourne-shell command-line parser; ignores complex structures
// (such as for, while, case, if), parses out simple command/arg lines.
//
// This isn't really intended as a generic /bin/sh parser, it's aimed at
// building data for StartUp!.
// ----------------------------------------------------------------------

#include "CommandParser.h"

#include <support/Errors.h>
#include <support/List.h>
#include <support/String.h>
#include <ctype.h>
#include <string.h>

#include <stdio.h>

// ----------------------------------------------------------------------
// Constructor/destructor
CommandParser::CommandParser( void )
	: _last_dir( NULL )
{
	_state.push( CMD_CLEAN );
}
	  
CommandParser::~CommandParser( void )
{
	if( _last_dir ) delete _last_dir;
}

// ----------------------------------------------------------------------
// Attempt to parse a command-line
//
// Returns the "cmd" and "args" in out, "directory" if a "cd" or "pushd"
// occurred before.  Return values:
//
// B_OK			- finished parsing this statement
// B_ERROR		- some error occured, out doesn't have useful data
// B_BUSY		- line was continued
// B_CANCELED	- line was a comment or blank

status_t CommandParser::Parse( const char *line, BMessage &out, bool need_cmd )
{
	// Make sure we're clean.
	if( _state.top() == CMD_CLEAN && !out.IsEmpty() ) {
		if( out.MakeEmpty() != B_OK ) {
			return B_ERROR;
		}
	}

	// Bail early if we can.
	int line_len = strlen( line );
	if( line == NULL || line_len == 0 ) return B_ERROR;

	// Is this continued?
	if( _state.top() == CMD_CONTINUED ) _state.pop();

	// Have we already gotten a command?
	bool found_cmd = false;
	if( !need_cmd || out.HasString( "cmd" ) ) found_cmd = true;

	bool added_line = false;
	if( found_cmd ) {
		out.AddString( "orig_args", line );
		added_line = true;
	}

	// Eat whitespace at the front.
	int idx = 0;
	while( isspace( line[idx] ) && idx < line_len ) idx++;

	// If this is a comment, bail.
	if( line[idx] == '#' ) return B_CANCELED;

	// Look for tokens.
	bool done = false;
	
	escape_stack escapes;
	escapes.push( ' ' );	// space is not an escape

	while( !done && idx < line_len ) {
		BString token( "" );

		bool token_end = false;
		bool token_redirect = false;

		while( !token_end && idx < line_len ) {
			switch( line[idx] ) {
			case '\\':
				if( escapes.top() != '\\' ) {
					escapes.push( '\\' );
				} else {
					escapes.pop();
					token.Append( line[idx], 1 );
				}
				break;

			case '\'':
				if( escapes.top() == '\\' || escapes.top() == '"' ) {
					token.Append( '\'', 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				} else {
					if( escapes.top() != '\'' ) {
						escapes.push( '\'' );
					} else {
						escapes.pop();
					}
				}
				break;

			case '"':
				if( escapes.top() == '\\' || escapes.top() == '\'' ) {
					token.Append( '"', 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				} else {
					if( escapes.top() != '"' ) {
						escapes.push( '"' );
					} else {
						escapes.pop();
					}
				}
				break;

			case ' ':	// fall-through
			case '\t':
				if( escapes.top() == ' ' ) {
					token_end = true;
				} else {
					if( token.Length() > 0 || escapes.top() == '\\' ) {
						token.Append( line[idx], 1 );
					}
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '#':
				if( escapes.top() == ' ' ) {
					_state.push( CMD_IN_COMMENT );
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '=':
				if( escapes.top() == ' ' && !found_cmd ) {
					token_end = true;
					if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '(':
				if( escapes.top() == ' ' ) {
					_state.push( CMD_IN_SUBSHELL );
					if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case ')':
				if( _state.top() == CMD_IN_SUBSHELL ) {
					_state.pop();
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '{':
				if( escapes.top() == ' ' ) {
					_state.push( CMD_IN_FUNCTION );
					if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '}':
				if( _state.top() == CMD_IN_FUNCTION ) {
					_state.pop();
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '&':
				if( escapes.top() == ' ' && !token_redirect ) {
					if( !out.HasBool( "&" ) ) out.AddBool( "&", true );
				} else {
					token.Append( line[idx], 1 );
					if( escapes.top() == '\\' ) escapes.pop();
				}
				break;

			case '|':
				if( escapes.top() == ' ' ) {
					if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
				} else {
					if( escapes.top() == '\\' ) escapes.pop();
				}
				token.Append( line[idx], 1 );
				break;

			case '<':	// fall-through
			case '>':
				if( escapes.top() == ' ' ) token_redirect = true;
				if( escapes.top() == '\\' ) escapes.pop();
				token.Append( line[idx], 1 );
				break;
			
			case '\n':	// fall-through
			case '\r':
				// Ignore end-of-line characters
				break;

			default:
				token.Append( line[idx], 1 );
				if( escapes.top() == '\\' ) escapes.pop();
				break;
			}
			
			idx++;
		}

		if( !token.Compare( "if" ) ) {
			_state.push( CMD_IN_IF );
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		} else if( !token.Compare( "for" ) ) {
			_state.push( CMD_IN_FOR );
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		} else if( !token.Compare( "while" ) ) {
			_state.push( CMD_IN_WHILE );
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		} else if( !token.Compare( "case" ) ) {
			_state.push( CMD_IN_CASE );
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		} else if( !token.Compare( "fi" ) ) {
			if( _state.top() == CMD_IN_IF ) _state.pop();
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		} else if( !token.Compare( "done" ) ) {
			if( _state.top() == CMD_IN_FOR || _state.top() == CMD_IN_WHILE ) _state.pop();
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		} else if( !token.Compare( "esac" ) ) {
			if( _state.top() == CMD_IN_CASE ) _state.pop();
			if( !out.HasBool( "complex" ) ) out.AddBool( "complex", true );
		}

		if( _state.top() == CMD_SKIP_REDIRECT ) {
			out.AddString( "redirects", token.String() );
			_state.pop();
		} else if( !found_cmd && token.Length() > 0 ) {
			found_cmd = true;

			out.AddString( "cmd", token.String() );

			if( !added_line ) {
				out.AddString( "orig_args", &(line[idx]) );
				added_line = true;
			}
		} else {
			if( token.Length() > 0 ) {
				if( token_redirect ) {
					out.AddString( "redirects", token.String() );
					_state.push( CMD_SKIP_REDIRECT );
				} else {
					out.AddString( "args", token.String() );
				}
			}
		}
		
		token.SetTo( "" );
	}

	// End of line... if we're not currently in a comment, a \ as the last
	// character continues.
	if( escapes.top() == '\\' && _state.top() != CMD_IN_COMMENT ) {
		_state.push( CMD_CONTINUED );
		return B_BUSY;
	}

	// We'll consider any of these to be continued statements, since we're
	// not handling them.
	switch( _state.top() ) {
	case CMD_IN_IF:
	case CMD_IN_FOR:
	case CMD_IN_WHILE:
	case CMD_IN_CASE:
	case CMD_IN_SUBSHELL:
	case CMD_IN_FUNCTION:
		return B_BUSY;
		break;
	default:
		break;
	}

	return B_OK;
}
