// ----------------------------------------------------------------------
// StartUp! - Launch window
//
// Copyright © 2000 Arcane Dragon Software (arcanedragon@pobox.com)
//
// Display and manage the Launch window.
// ----------------------------------------------------------------------

#include "App.h"
#include "LaunchWindow.h"
#include "ItemIconView.h"
#include "MutexList.h"
#include "StartupItem.h"
#include "StartupSettings.h"
#include "SysLog.h"

#include <app/Application.h>
#include <game/FileGameSound.h>
#include <interface/Font.h>
#include <storage/Entry.h>
#include <storage/File.h>
#include <storage/Path.h>
#include <storage/Resources.h>
#include <storage/StorageDefs.h>
#include <support/Errors.h>
#include <stdio.h>
#include <string.h>

#include "HGroup.h"
#include "MBViewWrapper.h"
#include "MBorder.h"
#include "MProgressBar.h"
#include "MStringView.h"
#include "Space.h"
#include "VGroup.h"

#define NO_SFX

// +---------------+
// | [ ]  StartUp! |
// +---------------+----------------------+
// |                                      |
// |  Launching…                          |
// |                                      |
// |  +--+  ItemName              n of m  |
// |  |1 |  ######______________________  |
// |  +--+                                |
// |                                      |
// +--------------------------------------+

// ----------------------------------------------------------------------
// Local constants

static const char *SETTING_POSITION = "about:position";

// ----------------------------------------------------------------------
// Local functions

static int numlen( int x )
{
	// Return the number of digits in a number.
	int d = 1;
	while( ( x /= 10 ) != 0 ) d++;
	return d;
}

int32 launch_thread( void *data )
{
	LaunchWindow *win = static_cast<LaunchWindow *>( data );

	MutexList *items = win->Items();
	if( !items || items->CountItems() <= 0 ) return B_ERROR;

	// Loop through the items and launch them.
	int num_items = items->CountItems();
	if( items->Lock() ) {
		for( int i = 0; i < num_items; i++ ) {
			StartupItem *item = static_cast<StartupItem *>( items->ItemAt( i ) );

			win->UpdateDisplay( item, i, num_items );

			status_t retval = item->Launch();
			if( retval != B_OK ) {
				SysLog_Error( "syslog:error launching", retval );
			}

			// TODO: should we have a delay between items? or make it a
			//       setting on each item?
			//item->LaunchDelay(); ?
			//snooze( 1 * 1000 * 1000 );

			if( win->IsQuitting() ) break;
		}

		items->Unlock();
	}

	if( !win->IsQuitting() ) {
		win->Done();

		// Close our window
		win->PostMessage( B_QUIT_REQUESTED );
	}

	return B_OK;
}

// ----------------------------------------------------------------------
// Constructor/destructor

LaunchWindow::LaunchWindow( MutexList *items )
	: MWindow( BRect( 100, 100, 101, 101 ),
			   static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "launch title", &_size ) ),
			   B_TITLED_WINDOW_LOOK, B_FLOATING_ALL_WINDOW_FEEL,
			   B_NOT_RESIZABLE | B_NOT_ZOOMABLE |
//			   B_WILL_ACCEPT_FIRST_CLICK | B_ASYNCHRONOUS_CONTROLS ),
			   B_ASYNCHRONOUS_CONTROLS ),
	  _tid( -1 ),
	  _quitting( false ),
	  _items( items ),
	  _nofm_format( NULL ),
	  _action_view( NULL ),
	  _icon_view( NULL ),
	  _item_name_view( NULL ),
	  _nofm_view( NULL ),
	  _status_view( NULL ),
	  _sfx( NULL )
{
	App *app = static_cast<App *>( be_app );
	app->RegisterWindow( this );

	BResources *rsrc = be_app->AppResources();

	// TODO: Be nice when we can replace this with a "real" BRsrcGameSound
	//		 or something.
#ifndef NO_SFX
	const void *sfx_data = rsrc->LoadResource( B_RAW_TYPE, "liquid pouring", &_size );
	BPath sfx_path( "/tmp/StartUp!_liquid_pouring" );
	BEntry sfx_ent( sfx_path.Path() );
	BFile *sfx_file = new BFile( &sfx_ent, B_WRITE_ONLY | B_CREATE_FILE );
	sfx_file->Write( sfx_data, _size );
	delete sfx_file;

	_sfx = new BFileGameSound( sfx_path.Path(), false );
	if( _sfx->InitCheck() != B_OK ) {
		// ugh...
		delete _sfx;
		_sfx = NULL;
	} else {
		_sfx->Preload();
		_sfx->StartPlaying();
	}

	sfx_ent.Remove();
#endif

	const char *launch_action = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "launch launching", &_size ) );
	_nofm_format = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "launch n of m", &_size ) );

	// Find the longest value for the item name string.
	float longest_item_name = 0.0;

	int num_items = items->CountItems();
	const char **item_names = new const char *[num_items];
	int32 *item_lengths = new int32[num_items];
	float *item_string_sizes = new float[num_items];
	int i = 0;	// Sure wish gcc would figure out scoping rules properly.
	for( /* int */ i = 0; i < num_items; i++ ) {
		StartupItem *item = static_cast<StartupItem *>( items->ItemAt( i ) );
		item_names[i] = item->GetName();
		item_lengths[i] = strlen( item_names[i] );
		item_string_sizes[i] = 0.0;
	}

	be_plain_font->GetStringWidths( item_names, item_lengths, num_items, item_string_sizes );
	for( /* int */ i = 0; i < num_items; i++ ) {
		if( item_string_sizes[i] > longest_item_name ) longest_item_name = item_string_sizes[i];
	}

	delete [] item_names;
	delete [] item_lengths;
	delete [] item_string_sizes;

	// Find the max length of the nofm view.
	float nofm_length = 0.0;

	const char *digits = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "digits", &_size ) );
	float longest_digit_size = 0;
	char longest_digit = digits[0];
	for( /* int */ i = 0; i < static_cast<int>( strlen( digits ) ); i++ ) {
		float width = be_plain_font->StringWidth( &(digits[i]), 1 );
		if( width > longest_digit_size ) {
			longest_digit_size = width;
			longest_digit = digits[i];
		}
	}

	int nofm_bufflen = 2 * numlen( num_items ) + strlen( _nofm_format );
	char *nofm_buff = new char[nofm_bufflen];	// remember to delete [] me!
	int nofm_max = longest_digit;
	for( /* int */ i = 1; i < num_items; i++ ) {
		nofm_max *= 10;
		nofm_max += longest_digit;
	}
	snprintf( nofm_buff, nofm_bufflen,
			  _nofm_format, nofm_max, nofm_max );

	nofm_length = be_plain_font->StringWidth( nofm_buff );

	// View hierarchy:
	//
	// MBorder
	// + VGroup
	//   + MStringView - Launching…
	//   + Space
	//   + HGroup
	//     + MBViewWrapper( ItemIconView )
	//     + Space
	//     + VGroup
	//       + HGroup
	//         + MStringView - item name
	//         + Space
	//         + MStringView - n of m
	//       + MBorder
	//         + MProgressBar

	float plain_height = 0.0;
	font_height h;
	be_plain_font->GetHeight( &h );
	plain_height = h.ascent + h.descent + h.leading;

	snprintf( nofm_buff, nofm_bufflen, _nofm_format, 0, num_items );

	MBorder *view = new MBorder(
						M_RAISED_BORDER, 16, "",
						new VGroup(
							( _action_view = new MStringView( launch_action ) ),

							new Space( minimax( 0, 16, 65536, 16 ) ),
							
							new HGroup(
								new MBViewWrapper( ( _icon_view = new ItemIconView ) ),
								new Space( minimax( 16, 0, 16, 65536 ) ),
								new VGroup(
									new HGroup(
										( _item_name_view = new MStringView( "", B_ALIGN_LEFT, minimax( longest_item_name, plain_height, 65536, 65536 ) ) ),
										new Space( minimax( 16, 0, 65536, 65536 ) ),
										( _nofm_view = new MStringView( nofm_buff, B_ALIGN_RIGHT, minimax( nofm_length, plain_height, 65536, 65536 ) ) ),
										NULL
									),
									new Space( minimax( 0, 4, 65536, 4 ) ),
									new MBorder(
										M_DEPRESSED_BORDER, 1, "",
										( _status_view = new MProgressBar( this ) )
									),
									NULL
								),
								NULL
							),
							NULL
						)
					);

	AddChild( view );

	_status_view->ct_mpm = minimax( 200, 16, 65536, 65536 );
	_status_view->value = 0.0;

	BPoint pos;
	StartupSettings *settings = app->GetSettings();
	if( settings->FindPoint( SETTING_POSITION, &pos ) == B_OK )  {
		if( pos.x != 0.0 && pos.y != 0.0 ) MoveTo( pos );
	}
	
	Show();

	_tid = LaunchThread();

	delete [] nofm_buff;
}

LaunchWindow::~LaunchWindow( void )
{
#ifndef NO_SFX
	if( _sfx ) delete _sfx;
#endif
}

// ----------------------------------------------------------------------
// BWindow methods

bool LaunchWindow::QuitRequested( void )
{
	// Wait for our thread to exit.
	_quitting = true;
	status_t exit_val = B_OK;
	status_t retval = wait_for_thread( _tid, &exit_val );

	// We don't really care how it exitted at this point...
	retval = retval, exit_val = exit_val;

	App *app = static_cast<App *>( be_app );
	StartupSettings *settings = app->GetSettings();
	settings->SavePoint( SETTING_POSITION, Frame().LeftTop() );

	app->DeregisterWindow( this );

	return true;
}

// ----------------------------------------------------------------------
// LaunchWindow methods

void LaunchWindow::UpdateDisplay( StartupItem *item, int n, int m )
{
	// m is the total number of items, n is range 0 - (m - 1)... adjust
	// for range 1 - m
	n++;

	// update the display before we launch this item
	BBitmap *bits = item->GetIcon( B_LARGE_ICON );
	const char *name = item->GetName();

	int nofm_bufflen = 2 * numlen( m ) + strlen( _nofm_format );
	char *nofm_buff = new char[nofm_bufflen];	// remember to delete [] me!
	snprintf( nofm_buff, nofm_bufflen, _nofm_format, n, m );

	if( LockLooper() ) {
		if( bits ) _icon_view->SetIcon( bits );
		if( name ) _item_name_view->SetText( name );
		_nofm_view->SetText( nofm_buff );
		_status_view->SetValue( (float)n / (float)m );
		UnlockLooper();
	}
	
	delete [] nofm_buff;
}

void LaunchWindow::Done( void )
{
	// update the display because we're done
	BResources *rsrc = be_app->AppResources();
	const char *action = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "launch done", &_size ) );

	if( LockLooper() ) {
		_action_view->SetText( action );
		UnlockLooper();
	}

	// TODO: should make this delay a setting
#ifndef NO_SFX
	_sfx->SetGain( 0.0, 1 * 1000 * 1000 );
#endif
	snooze( 1 * 1000 * 1000 );

#ifndef NO_SFX
	if( _sfx && _sfx->IsPlaying() ) _sfx->StopPlaying();
#endif
}

// ----------------------------------------------------------------------
// Private methods

thread_id LaunchWindow::LaunchThread( void )
{
	BResources *rsrc = be_app->AppResources();
	const char *thread_name = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "launch thread name", &_size ) );
	int32 *thread_prio = static_cast<int32 *>( const_cast<void *>( rsrc->LoadResource( B_INT32_TYPE, "launch thread priority", &_size ) ) );

	thread_id tid = spawn_thread( launch_thread, thread_name, *thread_prio, this );
	if( tid < 0 ) return tid;	// error!
	
	status_t retval = resume_thread( tid );
	if( retval != B_OK ) return retval;	// error!

	return tid;
}
