// ----------------------------------------------------------------------
// StartUp! - main window
//
// Copyright © 2000 Arcane Dragon Software (arcanedragon@pobox.com)
//
// Display and manage the main interface window.
// ----------------------------------------------------------------------

#include "App.h"
#include "AppIconMenu.h"
#include "CLVColumn.h"
#include "MainWindow.h"
#include "MutexList.h"
#include "NagWindow.h"
#include "StartupColumnList.h"
#include "StartupItem.h"
#include "StartupSettings.h"

#include <app/AppDefs.h>
#include <app/Messenger.h>
#include <interface/Menu.h>
#include <interface/MenuBar.h>
#include <interface/MenuItem.h>
#include <interface/TextView.h>
#include <storage/Directory.h>
#include <storage/Entry.h>
#include <storage/FilePanel.h>
#include <storage/Mime.h>
#include <storage/Node.h>
#include <storage/NodeMonitor.h>
#include <storage/Path.h>
#include <storage/Resources.h>
#include <stdio.h>

#include "HGroup.h"
#include "MBViewWrapper.h"
#include "MBorder.h"
#include "MButton.h"
#include "MMenuBar.h"
#include "MStringView.h"
#include "MTextControl.h"
#include "Space.h"
#include "VGroup.h"

// Original window design, with direct item editing in the items:
//
// +------------------------+
// | [ ]  StartUp! Items  # |
// +------------------------+--------------------------+
// | [1] Edit                                          |
// +---------------------------------------------------+
// |CLV of StartupItem objects; edit them directly here|
// +---------------------------------------------------+

// Temporary window design, with an item editing area.
//
// +------------------------+
// | [ ]  StartUp! Items  # |
// +------------------------+--------------------------+
// | [1] Edit                                          |
// +---------------------------------------------------+
// |                                                   |
// |  +-------------------------------------------+-+  |
// |  | CLV                                       | |  |
// |  |                                           | |  |
// |  |                                           | |  |
// |  |                                           | |  |
// |  +-------------------------------------------+-+  |
// |  |                                           | |  |
// |  +-------------------------------------------+-+  |
// |                                                   |
// |       Name: ____________________________________  |
// |  Arguments: ____________________________________  |
// |  Directory: (default)                             |
// |                                                   |
// +---------------------------------------------------+

// ----------------------------------------------------------------------
// Local constants

static const char *SETTING_POSITION = "main:position";

// ----------------------------------------------------------------------
// Constructor/destructor

MainWindow::MainWindow( MutexList *items )
	: MWindow( BRect( 100, 100, 599, 299 ),
			   static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "main title", &_size ) ),
//			   B_DOCUMENT_WINDOW_LOOK, B_NORMAL_WINDOW_FEEL,
			   B_TITLED_WINDOW_LOOK, B_NORMAL_WINDOW_FEEL,
//			   B_WILL_ACCEPT_FIRST_CLICK | B_ASYNCHRONOUS_CONTROLS ),
			   B_ASYNCHRONOUS_CONTROLS ),
	  _items( items ),
	  _item_view( NULL ),
	  _item_view_scroller( NULL ),
	  _name_field( NULL ),
	  _args_field( NULL ),
	  _dir_field( NULL ),
	  _current_edit( NULL ),
	  _directory_panel( NULL )
{
	App *app = static_cast<App *>( be_app );
	app->RegisterWindow( this );
	
	BResources *rsrc = be_app->AppResources();

	// Load strings from resources
	const char *app_about = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:app:about", &_size ) );
	const char *app_ads = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:app:ADS", &_size ) );
	const char *app_help = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:app:help", &_size ) );
	const char *app_reg = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "reg:title", &_size ) );
	const char *app_quit = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:app:quit", &_size ) );

	const char *edit_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:edit:title", &_size ) );
	const char *edit_cut = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:edit:cut", &_size ) );
	const char *edit_copy = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:edit:copy", &_size ) );
	const char *edit_paste = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:edit:paste", &_size ) );
	const char *edit_select_all = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "menu:edit:select all", &_size ) );

	char *name_label = const_cast<char *>( static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "name label", &_size ) ) );
	char *args_label = const_cast<char *>( static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "args label", &_size ) ) );
	char *dir_label = const_cast<char *>( static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "dir label", &_size ) ) );

	// Add the menu.
	MMenuBar *bar = new MMenuBar;
	bar->flags |= M_NO_Y_LEFTOVERS;
	
	BMenu *menu = new BMenu( "StartUp" );

		BMenuItem *item = new BMenuItem( app_about, new BMessage( B_ABOUT_REQUESTED ) );
		item->SetTarget( be_app );
		menu->AddItem( item );

		item = new BMenuItem( app_ads, new BMessage( App_SHOW_HOMEPAGE ) );
		item->SetTarget( be_app );
		menu->AddItem( item );
		
		menu->AddSeparatorItem();
		
		item = new BMenuItem( app_help, new BMessage( App_SHOW_HELP ) );
		item->SetTarget( be_app );
		menu->AddItem( item );
		
		item = new BMenuItem( app_reg, new BMessage( NAG_ENTER_KEY ) );
		item->SetTarget( be_app );
//		item->SetEnabled( !app->CheckKey() );
		item->SetEnabled( false );
		menu->AddItem( item );
		
		menu->AddSeparatorItem();

		item = new BMenuItem( app_quit, new BMessage( B_QUIT_REQUESTED ) );
		item->SetTarget( be_app );
		menu->AddItem( item );

	AppIconMenu *appicon = new AppIconMenu( menu );
	bar->AddItem( appicon );

	menu = new BMenu( edit_title );		// Edit menu
	
		item = new BMenuItem( edit_cut, new BMessage( B_CUT ) );
		item->SetTarget( this );
		item->SetEnabled( false );
		menu->AddItem( item );

		item = new BMenuItem( edit_copy, new BMessage( B_COPY ) );
		item->SetTarget( this );
		item->SetEnabled( false );
		menu->AddItem( item );

		item = new BMenuItem( edit_paste, new BMessage( B_PASTE ) );
		item->SetTarget( this );
		item->SetEnabled( false );
		menu->AddItem( item );

		menu->AddSeparatorItem();

		item = new BMenuItem( edit_select_all, new BMessage( B_SELECT_ALL ) );
		item->SetTarget( this );
		item->SetEnabled( false );
		menu->AddItem( item );

	bar->AddItem( menu );

	// Now add the column list view.

	const char *col_item_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "column:item", &_size ) );
	const char *col_args_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "column:args", &_size ) );
	const char *col_dir_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "column:dir", &_size ) );
//	const char *col_workspace_title = static_cast<const char *>( rsrc->LoadResource( B_STRING_TYPE, "column:workspace", &_size ) );

 	StartupSettings *settings = app->GetSettings();

	float col_item_width = 0.0;
	float col_args_width = 0.0;
	float col_dir_width = 0.0;
//	float col_workspace_width = 0.0;

	if( settings->FindFloat( "col:item:width", &col_item_width ) != B_OK ) col_item_width = 160.0;
	if( settings->FindFloat( "col:args:width", &col_args_width ) != B_OK ) col_args_width = 160.0;
	if( settings->FindFloat( "col:dir:width", &col_dir_width ) != B_OK ) col_dir_width = 160.0;
//	if( settings->FindFloat( "col:workspace:width", &col_workspace_width ) != B_OK ) col_workspace_width = 160.0;

	BRect clv_bounds = Bounds();
	clv_bounds.top = bar->Bounds().bottom + 1.0;	// argh...

	clv_bounds.SetRightBottom( clv_bounds.RightBottom() - 
				BPoint( B_H_SCROLL_BAR_HEIGHT, B_V_SCROLL_BAR_WIDTH ) );

	_item_view = new StartupColumnList( clv_bounds, &_item_view_scroller );

	_item_view->AddColumn( new CLVColumn( "", B_LARGE_ICON + 4.0, CLV_LOCK_AT_BEGINNING | CLV_NOT_MOVABLE | CLV_NOT_RESIZABLE, B_LARGE_ICON + 4.0 ) );
	_item_view->AddColumn( new CLVColumn( col_item_title, col_item_width, CLV_SORT_KEYABLE | CLV_TELL_ITEMS_WIDTH | CLV_HEADER_TRUNCATE ) );
	_item_view->AddColumn( new CLVColumn( col_args_title, col_args_width, CLV_SORT_KEYABLE | CLV_TELL_ITEMS_WIDTH | CLV_HEADER_TRUNCATE ) );
	_item_view->AddColumn( new CLVColumn( col_dir_title, col_dir_width, CLV_SORT_KEYABLE | CLV_TELL_ITEMS_WIDTH | CLV_HEADER_TRUNCATE ) );
//	_item_view->AddColumn( new CLVColumn( col_workspace_title, col_workspace_width, CLV_SORT_KEYABLE | CLV_TELL_ITEMS_WIDTH | CLV_HEADER_TRUNCATE ) );

	// MWindow
	// + MBorder
	//   + VGroup
	//     + MBViewWrapper( bar )
	//     + MBorder
	//       + VGroup
	//       + MBViewWrapper( CLV )
	//       + Space
	//       + MTextControl
	//       + MTextControl
	//       + MTextControl

	MBorder *view = new MBorder(
							M_NO_BORDER, 0, "",
							new VGroup(
								bar,

								new MBorder(
									M_RAISED_BORDER, 16, "",
									new VGroup(
										new MBViewWrapper( _item_view_scroller, false, false, false ),

										new Space( minimax( 0, 16, 65536, 16 ) ),
										
										( _name_field = new MTextControl( name_label, NULL, new BMessage( MAIN_EDITED_NAME ) ) ),
										( _args_field = new MTextControl( args_label, NULL, new BMessage( MAIN_EDITED_ARGS ) ) ),
										( _dir_field = new MTextControl( dir_label, NULL, new BMessage( MAIN_EDITED_DIRECTORY ) ) ),
										
										NULL
									)
								),
								NULL
							)
						);

	AddChild( view );

	// Make all the fields align properly.
	_name_field->SetAlignment( B_ALIGN_RIGHT, B_ALIGN_LEFT );
	_args_field->SetAlignment( B_ALIGN_RIGHT, B_ALIGN_LEFT );
	_dir_field->SetAlignment( B_ALIGN_RIGHT, B_ALIGN_LEFT );

	// Set the modification messages for the fields.
	_name_field->SetModificationMessage( new BMessage( MAIN_MODIFIED_NAME ) );
	_args_field->SetModificationMessage( new BMessage( MAIN_MODIFIED_ARGS ) );
	_dir_field->SetModificationMessage( new BMessage( MAIN_MODIFIED_DIRECTORY ) );

	// Make sure the fields aren't stylable.
	_name_field->TextView()->SetStylable( false );
	_args_field->TextView()->SetStylable( false );
	_dir_field->TextView()->SetStylable( false );
	
	DivideSame( _name_field, _args_field, _dir_field, NULL );

	type_code type;
	int32 count;
	if( settings->GetInfo( "col:order", &type, &count ) == B_OK ) {
		if( type == B_INT32_TYPE && count > 0 ) {
			if( count == COL_COUNT ) {
				int32 *order = new int32[count];
				for( int32 i = 0; i < count; i++ ) {
					(void)settings->FindInt32( "col:order", i, &(order[i]) );
				}
				_item_view->SetDisplayOrder( order );
				delete [] order;
			} else {
				(void)settings->RemoveName( "col:order" );
			}
		}
	}

	int32 sort_key = COL_NAME;
//	if( settings->FindInt32( "col:key", &sort_key ) ) ;	// do nothing
	_item_view->SetSortKey( sort_key );

	if( _items->Lock() ) {
		if( !_item_view->AddList( items ) ) {
			// that's bad...
		}
		_items->Unlock();
	}

	BRect pos;
	if( settings->FindRect( SETTING_POSITION, &pos ) == B_OK ) {
		if( pos.left != 0.0 && pos.top != 0.0 ) MoveTo( pos.LeftTop() );
		if( pos.Width() != 0.0 && pos.Height() != 0.0 ) ResizeTo( pos.Width(), pos.Height() );
	}

	// Need to add a node monitor on the Startup Items directory so
	// we can pick up any new items that appear from other sources.
	BPath *path = app->GetStartupPath();
	BDirectory dir( path->Path() );
	node_ref node;
	if( dir.GetNodeRef( &node ) == B_OK ) {
		status_t retval = watch_node( &node, B_WATCH_DIRECTORY, this );
		if( retval != B_OK ) {
			// waaah!
		}
	}

	Show();
}

MainWindow::~MainWindow( void )
{
	// Need to remove our node monitor
	stop_watching( this );

	if( _directory_panel ) delete _directory_panel;
}

// ----------------------------------------------------------------------
// BWindow methods

void MainWindow::MessageReceived( BMessage *msg )
{
	switch( msg->what ) {
	case B_SIMPLE_DATA:
		msg->what = B_REFS_RECEIVED;
		be_app->PostMessage( msg );
		break;

	case B_CUT:
	case B_COPY:
	case B_PASTE:
	case B_SELECT_ALL:
		// TODO: implement these
		break;

	case B_NODE_MONITOR:
		{
			int32 op;
			if( msg->FindInt32( "opcode", &op ) != B_OK ) break;
			
			switch( op ) {
			case B_ENTRY_CREATED:
				// new Startup Item is here
				{
					entry_ref ref;
					const char *name;

					if( msg->FindInt32( "device", &ref.device ) != B_OK ) break;
					if( msg->FindInt64( "directory", &ref.directory ) != B_OK ) break;
					if( msg->FindString( "name", &name ) != B_OK ) break;

					ref.set_name( name );
					
					// TODO: add to Startup Items list
					// TODO: add to _item_view
					BEntry ent( &ref );
					if( ent.InitCheck() != B_OK ) break;

					StartupItem *item = new StartupItem( &ent );
					if( item->InitCheck() != B_OK ) {
						delete item;
						break;
					}
					
					if( _items->Lock() ) {
						_items->AddItem( item );
						_item_view->AddItem( item );
						_item_view->SortItems();
						_items->Unlock();
					}
				}
				break;
				
			case B_ENTRY_REMOVED:
				// Startup Item removed from the Startup Items directory
				
				// TODO:
				// FUCK, I hate the storage kit.  We have a node_ref here,
				// and no way to get back to a BEntry that we put into the
				// Startup Items list.  So we'll have to add some method of
				// identifying them via their node_ref.  Bleh.
				{
					node_ref noder;
					if( msg->FindInt32( "device", &noder.device ) != B_OK ) break;
					if( msg->FindInt64( "node", &noder.node ) != B_OK ) break;
					if( _items->Lock() ) {
						for( int32 i = 0; i < _items->CountItems(); i++ ) {
							StartupItem *item = static_cast<StartupItem *>( _items->ItemAt( i ) );
							if( item->CheckNode( noder ) ) {
								_items->RemoveItem( item );
								_item_view->RemoveItem( item );
								break;
							}
						}
						_items->Unlock();
					}
				}
				break;
			
			case B_ENTRY_MOVED:
				// Item moved into this directory, or moved out of it
				
				// TODO:
				// Need to get the inode of the Startup Items directory to
				// see if we're adding or removing a file from our field of
				// view.  This sucks too.
				{
					// Find the Startup Items inode.
					App *app = static_cast<App *>( be_app );
					BPath *path = app->GetStartupPath();
					BEntry path_entry( path->Path(), true );
					if( path_entry.InitCheck() != B_OK ) break;
					node_ref path_node;
					if( path_entry.GetNodeRef( &path_node ) != B_OK ) break;

					// Find the moved item's info.
					entry_ref ref;
					const char *name;

					if( msg->FindInt32( "device", &ref.device ) != B_OK ) break;
					if( msg->FindInt64( "to directory", &ref.directory ) != B_OK ) break;
					if( msg->FindString( "name", &name ) != B_OK ) break;

					ref.set_name( name );

					// Are we moving in or moving out?
					if( ref.directory == path_node.node ) {
						// Moving in

						// Get an entry for this item
						BEntry ent( &ref );
						if( ent.InitCheck() != B_OK ) break;

						// Add the item; duplicated code sucks!
						StartupItem *item = new StartupItem( &ent );
						if( item->InitCheck() != B_OK ) {
							delete item;
							break;
						}
						
						if( _items->Lock() ) {
							_items->AddItem( item );
							_item_view->AddItem( item );
							_item_view->SortItems();
							_items->Unlock();
						}
					} else {
						// Moving out

						// Remove the item; duplicated code sucks!
						node_ref noder;
						if( msg->FindInt32( "device", &noder.device ) != B_OK ) break;
						if( msg->FindInt64( "node", &noder.node ) != B_OK ) break;
						if( _items->Lock() ) {
							for( int32 i = 0; i < _items->CountItems(); i++ ) {
								StartupItem *item = static_cast<StartupItem *>( _items->ItemAt( i ) );
								if( item->CheckNode( noder ) ) {
									_items->RemoveItem( item );
									_item_view->RemoveItem( item );
									break;
								}
							}
							_items->Unlock();
						}
					}
				}
				break;

			default:
				// don't care
				break;
			}
		}
		break;

	case ITEM_INVOKED:
		// ITEM_INVOKED could launch the item instead...
		{
			int32 index;
			if( msg->FindInt32( "index", &index ) == B_OK ) {
				StartupItem *item = static_cast<StartupItem *>( _item_view->ItemAt( index ) );
				if( item ) item->Launch();
			}
		}
		break;

	case ITEM_SELECTED:
		{
			int32 index;
			if( msg->FindInt32( "index", &index ) == B_OK ) {
				StartupItem *item = static_cast<StartupItem *>( _item_view->ItemAt( index ) );
				if( item ) EditItem( item );
			}
		}
		break;

	case MAIN_EDITED_NAME:	// fall through
	case MAIN_EDITED_ARGS:
	case MAIN_EDITED_DIRECTORY:
		if( _current_edit ) UpdateItem( _current_edit );
		break;

#if 0
	case MAIN_EDITED_DIRECTORY:
		if( _current_edit ) {
			// TODO: display the file selector
			SelectDirectory( _current_edit );

			//UpdateItem( _current_edit );
		}
		break;
#endif

	case MAIN_MODIFIED_NAME:	// fall through
	case MAIN_MODIFIED_ARGS:
	case MAIN_MODIFIED_DIRECTORY:
		if( _current_edit ) UpdateItem( _current_edit );
		break;


	// Messages from the file panel
	case B_CANCEL:
		// Who cares?
		break;

	case MAIN_DIRECTORY_CHOSEN:
		// Need to update an item!
		{
			StartupItem *item = NULL;
			if( msg->FindPointer( "item", (void **)( &item ) ) == B_OK ) {
				// Convert the chosen directory into a string.
				entry_ref ref;
				if( msg->FindRef( "refs", &ref ) == B_OK ) {
					BEntry entry( &ref, false );
					if( entry.InitCheck() == B_OK ) {
						BPath path( &ref );
						if( path.InitCheck() == B_OK ) {
							_dir_field->SetText( path.Path() );
						}
					}
				}
			
				UpdateItem( item );
			}
		}
		break;

	default:
		BWindow::MessageReceived( msg );
		break;
	}
}

bool MainWindow::QuitRequested( void )
{
	App *app = static_cast<App *>( be_app );
	StartupSettings *settings = app->GetSettings();

	if( LockLooper() ) {
		settings->SaveRect( SETTING_POSITION, Frame() );

		(void)settings->RemoveName( "col:item:width" );
		(void)settings->RemoveName( "col:args:width" );
		(void)settings->RemoveName( "col:dir:width" );
//		(void)settings->RemoveName( "col:workspace:width" );
		(void)settings->AddFloat( "col:item:width", _item_view->ColumnAt( COL_NAME )->Width() );
		(void)settings->AddFloat( "col:args:width", _item_view->ColumnAt( COL_ARGS )->Width() );
		(void)settings->AddFloat( "col:dir:width", _item_view->ColumnAt( COL_DIRECTORY )->Width() );
//		(void)settings->AddFloat( "col:workspace:width", _item_view->ColumnAt( COL_WORKSPACE )->Width() );

		int32 *order = new int32[_item_view->CountColumns()];
		_item_view->GetDisplayOrder( order );
		(void)settings->RemoveName( "col:order" );
		for( int32 i = 0; i < _item_view->CountColumns(); i++ ) {
			(void)settings->AddInt32( "col:order", order[i] );
		}

// oh, wait, this isn't trivial...
//		(void)settings->RemoveName( "col:key" );
//		(void)settings->AddInt32( "col:key", _item_view->

		UnlockLooper();
	}

	app->DeregisterWindow( this );

	return true;
}

// ----------------------------------------------------------------------
// MainWindow methods

void MainWindow::EditItem( StartupItem *item )
{
	if( LockLooper() ) {
		if( _current_edit ) UpdateItem( _current_edit );
	
		_current_edit = item;

		_name_field->SetText( item->GetName() );
		_args_field->SetText( item->GetArgs() );
		_dir_field->SetText( item->GetDirectory() );

		UnlockLooper();
	}
}

void MainWindow::SelectDirectory( StartupItem *item )
{
	if( !_directory_panel ) {
		// Create a new "open" panel
		BMessenger target( this );

		_directory_panel = new BFilePanel( B_OPEN_PANEL,
										   &target,
										   NULL,
										   B_DIRECTORY_NODE,
										   false );	// no multi-select
	}

	// Find this item's directory, if it has one.
	const char *dir = item->GetDirectory();
	if( dir ) {
		BPath path( dir );
		if( path.InitCheck() == B_OK ) {
			if( path.GetParent( &path ) != B_OK ) {
				path.SetTo( dir );
			}

			_directory_panel->SetPanelDirectory( dir );
		}
	}

	// Make a new message for us.
	BMessage notification( MAIN_DIRECTORY_CHOSEN );
	notification.AddPointer( "item", item );
	_directory_panel->SetMessage( &notification );

	// Go!
	_directory_panel->Show();
}

void MainWindow::UpdateItem( StartupItem *item )
{
	if( !item->SetName( _name_field->Text() ) ) {
		// why did that fail?
	}
	
	if( !item->SetArgs( _args_field->Text() ) ) {
		// why did that fail?
	}
	
	if( !item->SetDirectory( _dir_field->Text() ) ) {
		// why did that fail?
	}

	_item_view->InvalidateItem( _item_view->IndexOf( item ) );
}
