// ----------------------------------------------------------------------
// StartupItem.cpp
//
// Copyright © 2000 Arcane Dragon Software (arcanedragon@pobox.com)
//
// An item in the StartUp! items folder; can tell you its arguments and
// entry_ref.
// ----------------------------------------------------------------------

#include "ColumnListView.h"
#include "Colors.h"
#include "CommandParser.h"
#include "StartupItem.h"
#include "SysLog.h"

#include <app/Application.h>
#include <app/Message.h>
#include <app/Roster.h>
#include <interface/Bitmap.h>
#include <interface/Font.h>
#include <interface/InterfaceDefs.h>
#include <interface/Point.h>
#include <interface/View.h>
#include <kernel/fs_attr.h>
#include <storage/Entry.h>
#include <storage/Node.h>
#include <storage/NodeInfo.h>
#include <storage/Resources.h>
#include <storage/StorageDefs.h>
#include <support/Errors.h>
#include <support/String.h>
#include <errno.h>
#include <limits.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

// ----------------------------------------------------------------------
// Constructor/destructor

// TODO: We might want to move to storing entry_refs insted of BEntrys,
// 		 since apparently BEntry wastes an FD on you.  Stupid non-dynamic
//		 app limits...
StartupItem::StartupItem( BEntry *entry )
	: CLVEasyItem( 0, false, false, 32.0 ),
	  _entry( NULL ),
	  _settings( NULL ),
	  _large_icon( NULL ),
	  _mini_icon( NULL ),
	  _name( NULL ),
	  _args( NULL ),
	  _directory( NULL ),
	  _argv( NULL ),
	  _status( B_OK )
{
	_entry = new BEntry( *entry );
	if( _entry->InitCheck() != B_OK ) {
		_status = _entry->InitCheck();
		return;
	}
	status_t retval = _entry->GetNodeRef( &_noder );
	if( retval != B_OK ) {
		_status = retval;
		return;
	}

	// Load this item's settings or create a blank settings container.
	if( !LoadSettings() ) {
		_status = B_NO_INIT;
		return;
	}

	// Fill in the ColumnListView goodies
	SetColumnContent( COL_ICON, GetIcon( B_LARGE_ICON ), 2.0, false, true );
	SetColumnContent( COL_NAME, GetName() );
	if( GetArgs() == NULL ) {
		size_t size;
		const char *none = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "(none)", &size ) );
		SetColumnContent( COL_ARGS, none );
	} else {
		SetColumnContent( COL_ARGS, GetArgs() );
	}
	if( GetDirectory() == NULL ) {
		size_t size;
		const char *def = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "(default)", &size ) );
		SetColumnContent( COL_DIRECTORY, def );
	} else {
		SetColumnContent( COL_DIRECTORY, GetDirectory() );
	}
//	SetColumnContent( COL_WORKSPACE, ... );
}

StartupItem::~StartupItem( void )
{
	if( _argv ) delete [] _argv;
	if( _name ) free( const_cast<char *>( _name ) );				// allocated via strdup()
	if( _args ) free( const_cast<char *>( _args ) );				// allocated via strdup()
	if( _directory ) free( const_cast<char *>( _directory ) );		// allocated via strdup()
	if( _mini_icon ) delete _mini_icon;
	if( _large_icon ) delete _large_icon;
	if( _settings ) {
		if( !SaveSettings() ) {
			// gripe
		}
		delete _settings;
	}
	if( _entry ) delete _entry;
}

// ----------------------------------------------------------------------
// StartupItem methods

status_t StartupItem::Launch( void )
{
	// Launch me!

	entry_ref ref;

	status_t retval = _entry->GetRef( &ref );
	if( retval != B_OK ) return retval;

	// If I have a directory, go there.
	const char *dir = GetDirectory();
	if( dir ) {
		// Is there a Storage Kit way of doing this?
		if( chdir( dir ) != EOK ) {
			SysLog_Error( "syslog:item chdir", static_cast<status_t>( errno ) );
		}
	}

	if( GetArgc() > 0 ) {
		// We have command-line args.
		retval = be_roster->Launch( &ref, GetArgc(), GetArgv() );
	} else {
		// Just launch us.
		retval = be_roster->Launch( &ref );
	}

	if( retval == B_LAUNCH_FAILED_EXECUTABLE ) {
		// The file was marked as an executable (x-bit was set), but it
		// failed to run... we'll try to open it again.
		BEntry ent( &ref, true );
		BNode node( &ent );
		BNodeInfo nodeinfo( &node );
		if( nodeinfo.InitCheck() == B_OK ) {
			char buff[B_MIME_TYPE_LENGTH];
			if( nodeinfo.GetType( buff ) == B_OK ) {
				BMessage msg( B_REFS_RECEIVED );
				msg.AddRef( "refs", &ref );
				retval = be_roster->Launch( buff, &msg );
			}
		}
	}

	return retval;
}

BBitmap *StartupItem::GetIcon( icon_size which )
{
	// Get the item's icon bitmap.
	switch( which ) {
	case B_LARGE_ICON:
		if( _large_icon ) return _large_icon;
		break;
	case B_MINI_ICON:
		if( _mini_icon ) return _mini_icon;
		break;
	default:
		SysLog_Error( "syslog:invalid icon size" );
		return NULL;
		break;
	}

	entry_ref ref;
	status_t retval = _entry->GetRef( &ref );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no item ref", retval );
		return NULL;
	}

	BEntry ent( &ref, true );
	if( ent.InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no item entry", ent.InitCheck() );
		return NULL;
	}
	retval = ent.GetRef( &ref );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no item ref", retval );
		return NULL;
	}

	BBitmap *bits = new BBitmap( BRect( 0, 0, which - 1, which - 1 ), B_CMAP8 );
	retval = BNodeInfo::GetTrackerIcon( &ref, bits, which );
	if( retval != B_OK ) {
		delete bits;
		
		SysLog_Error( "syslog:no item icon", retval );
		return NULL;
	}
	
	switch( which ) {
		case B_LARGE_ICON:	_large_icon = bits;		break;
		case B_MINI_ICON:	_mini_icon = bits;		break;
		default:			/* won't get here */	break;
	}
	
	return bits;
}

const char *StartupItem::GetName( void )
{
	// Get the file name
	if( _name ) return _name;
	
	char *buff = new char[NAME_MAX];
	status_t retval = _entry->GetName( buff );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no item name", retval );
	} else {
		_name = strdup( buff );
	}

	delete [] buff;

	return _name;
}

const char *StartupItem::GetArgs( void )
{
	if( !_settings && !LoadSettings() ) return 0;

	if( _args ) return _args;
	
	size_t size;
	const char *argv_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:raw argv", &size ) );

	const char *raw_argv = NULL;
	status_t retval = _settings->FindString( argv_name, &raw_argv );
	if( retval != B_OK ) return NULL;
	
	return const_cast<const char *>( strdup( raw_argv ) );
}

bool StartupItem::CheckNode( node_ref &noder )
{
	if( noder.device == _noder.device &&
		noder.node == _noder.node     ) return true;

	return false;
}

int StartupItem::GetArgc( void )
{
	// Find the number of args in the settings
	if( !_settings && !LoadSettings() ) return 0;

	// If there are no argv strings in the settings...
	size_t size;
	const char *argv_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:argv", &size ) );

	type_code type;
	int32 count;
	status_t retval = _settings->GetInfo( argv_name, &type, &count );
	if( retval != B_OK ) return 0;
	
	if( type != B_STRING_TYPE ) {
		SysLog_Error( "syslog:bad settings args" );
		return 0;
	}

	return static_cast<int>( count );
}

char **StartupItem::GetArgv( void )
{
	// Load the argv strings from our settings
	if( _argv ) return _argv;

	if( !_settings && !LoadSettings() ) return NULL;

	int argc = GetArgc();
	if( argc == 0 ) return NULL;

	size_t size;
	const char *argv_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:argv", &size ) );

	_argv = new char*[argc];
	for( int i = 0; i < argc; i++ ) {
		status_t retval = _settings->FindString( argv_name, i, const_cast<const char **>( &(_argv[i]) ) );
		if( retval != B_OK ) {
			SysLog_Error( "syslog:bad settings argv item", retval );
		}
	}

	return _argv;
}

const char *StartupItem::GetDirectory( void )
{
	if( _directory ) return _directory;

	if( !_settings && !LoadSettings() ) return NULL;

	size_t size;
	const char *dir_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:directory", &size ) );

	const char *dir = NULL;
	if( _settings->FindString( dir_name, &dir ) == B_OK ) {
		_directory = strdup( dir );
	}

	return _directory;
}

bool StartupItem::SetArgs( const char *args )
{
	// Take a command-line, and split it into args.  Save each arg under
	// the "settings:argv" string, save the raw args string under
	// "settings:raw argv".

	if( _args ) free( const_cast<char *>( _args ) );
	if( _argv ) delete [] _argv;

	if( args == NULL || strlen( args ) == 0 ) {
		_args = NULL;

		size_t size;
		const char *argv_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:argv", &size ) );
		const char *argv_raw = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:raw argv", &size ) );
		_settings->RemoveName( argv_name );
		_settings->RemoveName( argv_raw );

		const char *none = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "(none)", &size ) );
		SetColumnContent( COL_ARGS, none );
	} else {
		CommandParser parser;
		BMessage argmsg;
		switch( parser.Parse( args, argmsg, false ) ) {
		case B_OK:
			// complete args are there
			break;
			
		case B_ERROR:
			// no good
			// TODO: highlight args column somehow
			break;
			
		case B_BUSY:
			// not done yet
			// TODO: highlight args column
			break;
		
		case B_CANCELED:
			// comment or blank
			{
				_args = NULL;
	
				size_t size;
				const char *none = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "(none)", &size ) );
				SetColumnContent( COL_ARGS, none );
				
				return true;
			}
			break;
			
		default:
			// you won't get here
			break;
		}

		_args = strdup( args );

		size_t size;
		const char *argv_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:argv", &size ) );
		const char *argv_raw = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:raw argv", &size ) );
		if( !_settings && !LoadSettings() ) {
			// That's bad!
			return false;
		}
		_settings->RemoveName( argv_name );
		_settings->RemoveName( argv_raw );

		const char *one_arg = NULL;
		int32 index = 0;
		while( argmsg.FindString( "args", index++, &one_arg ) == B_OK ) {
			_settings->AddString( argv_name, one_arg );
		}

		_settings->AddString( argv_raw, args );

		SetColumnContent( COL_ARGS, args );
	}
	
	return true;
}

bool StartupItem::SetDirectory( const char *dir )
{
	// Set the item's startup directory
	free( const_cast<char *>( _directory ) );

	if( dir == NULL || strlen( dir ) == 0 ) {
		_directory = NULL;

		size_t size;
		const char *dir_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:directory", &size ) );
		const char *def = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "(default)", &size ) );

		_settings->RemoveName( dir_name );

		SetColumnContent( COL_DIRECTORY, def );
	} else {
		_directory = strdup( dir );

		size_t size;
		const char *dir_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:directory", &size ) );
		_settings->RemoveName( dir_name );
		_settings->AddString( dir_name, dir );

		SetColumnContent( COL_DIRECTORY, dir );
	}
	
	return true;
}

bool StartupItem::SetName( const char *name )
{
	// Rename the item
	if( !name || strlen( name ) == 0 ) return false;

	if( _entry->Rename( name ) != B_OK ) return false;

	free( const_cast<char *>( _name ) );
	_name = strdup( name );

	SetColumnContent( COL_NAME, name );
	
	return true;
}

bool StartupItem::SetPrefs( BMessage& msg )
{
	const char *dir = NULL;
	if( msg.FindString( "directory", &dir ) == B_OK ) {
		if( !SetDirectory( dir ) ) return false;
	}

	type_code code;
	int32 count;
	if( msg.GetInfo( "args", &code, &count ) != B_OK ) return false;
	if( code != B_STRING_TYPE ) return false;

	// This is a little silly after parsing it all...
	BString cmd;
	bool has_cmd = false;
	if( msg.FindString( "cmd", &cmd ) == B_OK ) has_cmd = true;

	BString args( "" );

	BString line;
	int32 i = 0;
	while( msg.FindString( "orig_args", i, &line ) == B_OK ) {
		if( i > 0 ) args.Append( ' ', 1 );
		args.Append( line );
		i++;
	}

	if( !SetArgs( args.String() ) ) return false;
	
	return true;
}

// ----------------------------------------------------------------------
// Private StartupItem methods

bool StartupItem::LoadSettings( void )
{
	// Load/unflatten our settings; if we can't load the settings from
	// our attributes, we create a fresh set of settings.
	if( _settings ) return true;
	
	// Why does this fail?  We _have_ a valid entry.
	BNode node( _entry );
	if( node.InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no item node", node.InitCheck() );
		return false;
	}
	
	size_t size;
	const char *attr_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:attr name", &size ) );

	attr_info info;
	status_t retval = node.GetAttrInfo( attr_name, &info );
	if( retval == B_ENTRY_NOT_FOUND ) {
		// No settings present, make a new one.
		_settings = new BMessage;
		return true;
	} else if( retval != B_OK ) {
		SysLog_Error( "syslog:no item settings", retval );
		return false;
	} else if( info.type != B_RAW_TYPE ) {
		SysLog_Error( "syslog:weird item settings" );
		return false;
	}

	// Now that we've actually got some attributes, load the flattened
	// BMessage data.
	_settings = new BMessage;
	char *buff = new char[info.size];
	ssize_t in_bytes = node.ReadAttr( attr_name, info.type, 0, buff, info.size );
	if( in_bytes != info.size ) {
		SysLog_Error( "syslog:no read item settings" );
	} else {
		// Unflatten that data!
		retval = _settings->Unflatten( buff );
		if( retval != B_OK ) {
			SysLog_Error( "syslog:no unflatten item settings" );
			delete _settings;
			_settings = new BMessage;
		}
	}

	delete [] buff;
	
	return true;
}

bool StartupItem::SaveSettings( void )
{
	// Flatten/save the settings
	if( !_settings ) return false;

	// Find this item's node
	BNode node( _entry );
	if( node.InitCheck() != B_OK ) {
		SysLog_Error( "syslog:no item node", node.InitCheck() );
		return false;
	}
	
	size_t size;
	const char *attr_name = static_cast<const char *>( be_app->AppResources()->LoadResource( B_STRING_TYPE, "settings:attr name", &size ) );

	// Flatten the settings
	ssize_t flatsize = _settings->FlattenedSize();
	char *buff = new char[flatsize];
	status_t retval = _settings->Flatten( buff, flatsize );
	if( retval != B_OK ) {
		SysLog_Error( "syslog:no flatten item settings", retval );
		delete [] buff;
		return false;
	}

	ssize_t bytes = node.WriteAttr( attr_name, B_RAW_TYPE, 0, buff, flatsize );
	if( bytes != flatsize ) {
		SysLog_Error( "syslog:no write item settings", bytes );
		delete [] buff;
		return false;
	}
	
	delete [] buff;
	return true;
}
